#include "PlayerJohannBrehmer.h"
#include <iostream>
#include <string>
#include <cstdlib>
#include <algorithm>

using namespace std;



PlayerJohannBrehmer::PlayerJohannBrehmer() : myName("Johann Brehmer"){
	M = 100.;
	myMode = 4;
	myParam = 1.;
	myParam2 = 0.2;
	thisGameBalance = 0.;
	totalBalance = 0.;
	roundsPlayedSoFarAsA = 0;
	roundsPlayedSoFarAsB = 0;
}



PlayerJohannBrehmer::PlayerJohannBrehmer(int mode, double param, double param2) : myName("Johann Brehmer"){
	M = 100.;
	myMode = mode;
	myParam = param;
	myParam2 = param2;
	thisGameBalance = 0.;
	totalBalance = 0.;
	roundsPlayedSoFarAsA = 0;
	roundsPlayedSoFarAsB = 0;
	
}



PlayerJohannBrehmer::~PlayerJohannBrehmer() {
	
}



double PlayerJohannBrehmer::checkAmount(double amount) {
	if (amount < 0) {
		return 0;
	}
	if (amount > thisGameBalance) {
		return thisGameBalance;
	}
	return amount;
}



double PlayerJohannBrehmer::ReceiveAndForwardMoneyA(double received_amount) {
	M = received_amount;
	thisGameBalance = received_amount;
	double giveMoney = checkAmount(theVerySecretAlgorithm(true));
	thisGameBalance -= giveMoney;
	//cout << roundsPlayedSoFarAsA << endl;
	if(roundsPlayedSoFarAsA < maxRounds)
	  MoneySentAsA[roundsPlayedSoFarAsA] = giveMoney;
	return giveMoney;
}



double PlayerJohannBrehmer::ReceiveAndForwardMoneyB(double received_amount) {
	thisGameBalance = received_amount;
	if(roundsPlayedSoFarAsB < maxRounds)
	  MoneyGotAsB[roundsPlayedSoFarAsB] = received_amount;
	double giveMoney = checkAmount(theVerySecretAlgorithm(false));
	thisGameBalance -= giveMoney;
	//cout << roundsPlayedSoFarAsB << endl;
	if(roundsPlayedSoFarAsB < maxRounds)
	  MoneySentBackAsB[roundsPlayedSoFarAsB] = giveMoney;
	roundsPlayedSoFarAsB++;
	totalBalance += thisGameBalance;
	return giveMoney;
}



void PlayerJohannBrehmer::RecieveMoneyBack(double received_amount) {
	thisGameBalance += received_amount;
	if(roundsPlayedSoFarAsA < maxRounds)
	  MoneyGotBackAsA[roundsPlayedSoFarAsA] = received_amount;
	roundsPlayedSoFarAsA++;	
	totalBalance += thisGameBalance;
}



double PlayerJohannBrehmer::GetBalance() const {
	return totalBalance;
}



string PlayerJohannBrehmer::GetName() const {
	return myName;
}



double PlayerJohannBrehmer::EstimatedOpponentGenerosityAsA(bool IncludeFirstRound) {
	double basis = myParam / (roundsPlayedSoFarAsB > 5 ? (double) roundsPlayedSoFarAsB / 5. : 1.);
	double sum = 0.;
	double norm = 0.;
	for (int i = (IncludeFirstRound ? max(0, roundsPlayedSoFarAsB - 5) : max(1, roundsPlayedSoFarAsB - 5)); i < roundsPlayedSoFarAsB; i ++) {
		sum += MoneyGotAsB[i] / (k*M);
		norm += k*M;
	}
	if (roundsPlayedSoFarAsB >= 5 && sum <= 0)
		return 0.;
	if (norm + basis > 0 && sum <= norm)
		return (sum + basis) / (norm+basis);
	else
		return 1.;
}



double PlayerJohannBrehmer::EstimatedOpponentGenerosityAsB(bool IncludeFirstRound) {
	double basis = myParam2 / (roundsPlayedSoFarAsA > 5 ? (double) roundsPlayedSoFarAsA / 5. : 1.);
	double sum = 0.;
	double norm = 0.;
	for (int i = (IncludeFirstRound ? max(0, roundsPlayedSoFarAsB - 5) : max(1, roundsPlayedSoFarAsB - 5)); i < roundsPlayedSoFarAsA; i ++) {
	  if(i < maxRounds){
		if (MoneySentAsA[i] > 0) {
			sum += MoneySentAsA[i]/M * (MoneyGotBackAsA[i] / (MoneySentAsA[i] * k));
			norm += MoneySentAsA[i]/M;
		}
	  }
	}
	if (roundsPlayedSoFarAsA >= 5 && sum <= 0)
		return 0.;
	if (norm + basis > 0 && sum <= norm)
		return (sum + basis) / (norm+basis);
	else
		return 1.;
}


double PlayerJohannBrehmer::theVerySecretAlgorithm(bool IAmPlayerA) {
	if (myMode == 0) { // Greedy! Keep everything
		return 0;
	}
	if (myMode == 1) { // Random
		return ((double) rand() / (double) RAND_MAX)*thisGameBalance*(IAmPlayerA ? myParam : myParam2);
	}
	if (myMode == 2) { // Relative: Always give back a constant fraction of what you have
		return (IAmPlayerA ? myParam*thisGameBalance : myParam2*thisGameBalance);
	}
	if (myMode == 3) { // Mirror: Do as the other did last
		if ((IAmPlayerA && roundsPlayedSoFarAsB == 0) || ((!IAmPlayerA) && roundsPlayedSoFarAsA == 0)) { // First round
			return 0.;
		} else {
			if (IAmPlayerA) {
				return MoneyGotAsB[roundsPlayedSoFarAsB-1] / k; 
			} else {
				if (MoneySentAsA[roundsPlayedSoFarAsA-1] <= 0) {
					return 0.5 * thisGameBalance;
				} else {
					return (MoneyGotBackAsA[roundsPlayedSoFarAsA-1] / (k * MoneySentAsA[roundsPlayedSoFarAsA-1])) * thisGameBalance;
				}
			}
		}
	}
	if (myMode == 4) { // Optimal
		if (IAmPlayerA) {
			if (roundsPlayedSoFarAsA==0)
				return 0.;
			if (EstimatedOpponentGenerosityAsB(true) >= 0.5)
				return thisGameBalance;
			return EstimatedOpponentGenerosityAsA(true)*thisGameBalance;
		} else {
			if (roundsPlayedSoFarAsB == 0)
				return 0;
			return EstimatedOpponentGenerosityAsB(true)*thisGameBalance;
		}
	}
	if (myMode == 5) { // Generous, also in first round
		if (IAmPlayerA) {
			return EstimatedOpponentGenerosityAsA(true)*thisGameBalance;
		} else {
			return EstimatedOpponentGenerosityAsB(true)*thisGameBalance;
		}
	}
	if (myMode == 6) { // Fox, doesnt take into account first round
		if (IAmPlayerA) {
			if (roundsPlayedSoFarAsA==0)
				return 0.;
			if (EstimatedOpponentGenerosityAsB(false) >= 0.5)
				return thisGameBalance;
			return EstimatedOpponentGenerosityAsA(false)*thisGameBalance;
		} else {
			if (roundsPlayedSoFarAsB == 0)
				return 0;
			return EstimatedOpponentGenerosityAsB(false)*thisGameBalance;
		}
	}
	// This should never happen:
	return 0.;
}
